

class Messages:

    def write_message(self, message: str, skip_newline=False):

        if not skip_newline:
            print(message)
        else:
            print(message, end='')

    def write_warning(self, message: str):

        print('WARNING: {}'.format(message))

    def write_exception(self, message: str):

        print('EXCEPTION: {}'.format(message))



import os
import cmd2
from cmd2 import Fg, style, Cmd2ArgumentParser, with_argparser, with_category

from nucleus_driver import NucleusDriver


class App(cmd2.Cmd):

    CMD_CAT_CONNECTING = 'Connection'
    CMD_CAT_LOGGING = 'Logging'
    CMD_CAT_COMMAND = 'Command'
    CMD_CAT_FLASH = 'Flash'
    CMD_CAT_SYSLOG = 'Syslog'
    CMD_CAT_ASSERT = 'Assert'
    CMD_CAT_DOWNLOAD = 'Download'

    def __init__(self, nucleus_driver):
        super().__init__()

        self.nucleus_driver = nucleus_driver

        self.prompt = style('nucleus> ', fg=Fg.GREEN)

        intro = ["  _   _            _                  _____       _                \r\n",
                 " | \ | |          | |                |  __ \     (_)               \r\n",
                 " |  \| |_   _  ___| | ___ _   _ ___  | |  | |_ __ ___   _____ _ __ \r\n",
                 " | . ` | | | |/ __| |/ _ \ | | / __| | |  | | '__| \ \ / / _ \ '__|\r\n",
                 " | |\  | |_| | (__| |  __/ |_| \__ \ | |__| | |  | |\ V /  __/ |   \r\n",
                 " |_| \_|\__,_|\___|_|\___|\__,_|___/ |_____/|_|  |_| \_/ \___|_|   \r\n"]

        self.intro = style(''.join(intro), fg=Fg.BLUE, bold=True)

        del cmd2.Cmd.do_edit
        del cmd2.Cmd.do_alias
        del cmd2.Cmd.do_macro
        del cmd2.Cmd.do_run_pyscript
        del cmd2.Cmd.do_run_script
        #del cmd2.Cmd.do_set
        del cmd2.Cmd.do_shell
        del cmd2.Cmd.do_shortcuts

    connect_parser = Cmd2ArgumentParser(description='Connect to nucleus device')
    connect_parser.add_argument('connection_type', choices=['serial', 'tcp'], help='connect to nucleus device through serial or tcp')

    @with_argparser(connect_parser)
    @with_category(CMD_CAT_CONNECTING)
    def do_connect(self, connect_args):

        def serial_configuration() -> bool:

            port = self.nucleus_driver.connection.select_serial_port()

            if port is None:
                return False

            self.nucleus_driver.connection.set_serial_configuration(port=port)

            return True

        def tcp_configuration() -> bool:

            self.nucleus_driver.messages.write_message('\nConnect through TCP with host(hostname/ip) or serial number: ')
            self.nucleus_driver.messages.write_message('[0] Host')
            self.nucleus_driver.messages.write_message('[1] Serial_number')
            reply = input('Input integer value in range [0:1]: ')

            if reply == '0':
                host = input('\ntcp - host: ')

                if host == '':
                    return False

                self.nucleus_driver.connection.set_tcp_configuration(host=host)

            elif reply == '1':
                serial_number = input('\ntcp - serial number: ')

                try:
                    int(serial_number)
                    condition = self.nucleus_driver.connection.set_tcp_hostname_from_serial_number(serial_number=serial_number)
                except ValueError:
                    self.nucleus_driver.messages.write_warning('serial number is not integer, host name will not be set')
                    condition = False

                if not condition:
                    return condition

            else:
                self.nucleus_driver.messages.write_message('Invalid selection')
                return False

            return True

        config_status = False

        if connect_args.connection_type == 'serial':
            config_status = serial_configuration()

        elif connect_args.connection_type == 'tcp':
            config_status = tcp_configuration()

        if not config_status:
            self.nucleus_driver.messages.write_warning('Failed to set connection configuration')
            return

        if self.nucleus_driver.connect(connection_type=connect_args.connection_type):
            self.nucleus_driver.messages.write_message('\r\nSuccessfully connected to Nucleus device\r\n')
            self.nucleus_driver.messages.write_message('ID:    {}'.format(self.nucleus_driver.connection.nucleus_id))
            self.nucleus_driver.messages.write_message('GETFW: {}\r\n'.format(self.nucleus_driver.connection.firmware_version))
        else:
            self.nucleus_driver.messages.write_warning('\r\nFailed to connect to Nucleus device\r\n')

    disconnect_parser = Cmd2ArgumentParser(description='Disconnect from the nucleus device')

    @with_argparser(disconnect_parser)
    @with_category(CMD_CAT_CONNECTING)
    def do_disconnect(self, _):

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('No device connected')
            return

        if self.nucleus_driver.disconnect():
            self.nucleus_driver.messages.write_message('Disconnected from Nucleus device')
        else:
            self.nucleus_driver.messages.write_warning('Failed to disconnect from Nucleus Device')

    start_measurement_parser = Cmd2ArgumentParser(description='Start measurement')
    start_measurement_parser.add_argument('-p', '--path', help='Set path for log files', completer=cmd2.Cmd.path_complete)

    @with_argparser(start_measurement_parser)
    @with_category(CMD_CAT_LOGGING)
    def do_start(self, start_measurement_args):

        path = start_measurement_args.path

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('Nucleus not connected')
            return

        if self.nucleus_driver.parser.thread.is_alive():
            self.nucleus_driver.messages.write_message('Logging thread is already running')
            return

        if path is not None:
            self.nucleus_driver.set_log_path(path=path)

        self.nucleus_driver.start_logging()
        self.nucleus_driver.start_measurement()

    start_field_calibration_parser = Cmd2ArgumentParser(description='Start field calibration')
    start_field_calibration_parser.add_argument('-p', '--path', help='Set path for log files', completer=cmd2.Cmd.path_complete)

    @with_argparser(start_field_calibration_parser)
    @with_category(CMD_CAT_LOGGING)
    def do_fieldcal(self, start_field_calibration_args):

        path = start_field_calibration_args.path

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('Nucleus not connected')
            return

        if self.nucleus_driver.parser.thread.is_alive():
            self.nucleus_driver.messages.write_message('Logging thread is already running')
            return

        if path is not None:
            self.nucleus_driver.set_log_path(path=path)

        self.nucleus_driver.start_logging()
        self.nucleus_driver.start_fieldcal()

    stop_measurement_parser = Cmd2ArgumentParser(description='Stop measurement')

    @with_argparser(stop_measurement_parser)
    @with_category(CMD_CAT_LOGGING)
    def do_stop(self, _):

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('Nucleus not connected')
            return

        self.nucleus_driver.stop_logging()
        self.nucleus_driver.stop()

    command_parser = Cmd2ArgumentParser(description='Send command to nucleus')
    command_parser.add_argument('command', help='send command to nucleus device and print response')

    @with_argparser(command_parser)
    @with_category(CMD_CAT_COMMAND)
    def do_command(self, command_args):

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('Nucleus not connected')
            return

        if self.nucleus_driver.parser.thread.is_alive():
            self.nucleus_driver.messages.write_message('Can not send command to Nucleus while logging thread is running')
            return

        command = command_args.command

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('Nucleus not connected')
            return

        self.nucleus_driver.send_command(command)

    flash_firmware_parser = Cmd2ArgumentParser(description='Flash firmware')
    flash_firmware_parser.add_argument('path', help='Set flash file. Extension must be in [.bin, .ldr, .zip]', completer=cmd2.Cmd.path_complete)

    @with_argparser(flash_firmware_parser)
    @with_category(CMD_CAT_FLASH)
    def do_flash_firmware(self, set_flash_path_args):

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('Nucleus not connected')
            return

        if self.nucleus_driver.parser.thread.is_alive():
            self.nucleus_driver.messages.write_message('Can not flash Nucleus while logging thread is running')
            return

        path = set_flash_path_args.path

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('Nucleus not connected')
            return

        self.nucleus_driver.flash_firmware(path=path)

    ###########################################
    # Assert
    ###########################################

    download_asserts_parser = Cmd2ArgumentParser(description='Download asserts')
    download_asserts_parser.add_argument('-p', '--path', help='Set path for assert file download', completer=cmd2.Cmd.path_complete)

    @with_argparser(download_asserts_parser)
    @with_category(CMD_CAT_ASSERT)
    def do_assert_download(self, download_asserts_args):

        path = download_asserts_args.path

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('Nucleus not connected')
            return

        if self.nucleus_driver.parser.thread.is_alive():
            self.nucleus_driver.messages.write_message('Can not download asserts while logging thread is running')
            return

        if not self.nucleus_driver.assert_download(path=path):
            self.nucleus_driver.messages.write_warning('Did not receive any asserts from Nucleus')

    clear_asserts_parser = Cmd2ArgumentParser(description='Clear asserts from Nucleus device')

    @with_argparser(clear_asserts_parser)
    @with_category(CMD_CAT_ASSERT)
    def do_assert_clear(self, _):

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('Nucleus not connected')
            return

        if self.nucleus_driver.parser.thread.is_alive():
            self.nucleus_driver.messages.write_message('Can not clear asserts while logging thread is running')
            return

        if b'OK\r\n' in self.nucleus_driver.clear_assert():
            self.nucleus_driver.messages.write_message('Successfully cleared asserts')
        else:
            self.nucleus_driver.messages.write_warning('Clear assert did not reply with OK: {}'.format(response))

    ##########################################
    # Syslog
    ###########################################

    download_syslog_parser = Cmd2ArgumentParser(description='Download syslog')
    download_syslog_parser.add_argument('-p', '--path', help='Set path for syslog file download', completer=cmd2.Cmd.path_complete)

    @with_argparser(download_syslog_parser)
    @with_category(CMD_CAT_SYSLOG)
    def do_syslog_download(self, download_syslog_args):

        path = download_syslog_args.path

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('Nucleus not connected')
            return

        if self.nucleus_driver.parser.thread.is_alive():
            self.nucleus_driver.messages.write_message('Can not download syslog while logging thread is running')
            return

        if not self.nucleus_driver.syslog_download(path=path):
            self.nucleus_driver.messages.write_warning('Did not receive any syslog entries from Nucleus')

    clear_syslog_parser = Cmd2ArgumentParser(description='Clear syslog from Nucleus device')

    @with_argparser(clear_syslog_parser)
    @with_category(CMD_CAT_SYSLOG)
    def do_syslog_clear(self, _):

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('Nucleus not connected')
            return

        if self.nucleus_driver.parser.thread.is_alive():
            self.nucleus_driver.messages.write_message('Can not clear syslog while logging thread is running')
            return

        response = self.nucleus_driver.clear_syslog()

        for message in response:
            self.nucleus_driver.messages.write_message(message)

    ##########################################
    # Download
    ###########################################

    download_dvl_parser = Cmd2ArgumentParser(description='Download dvl diagnostics data')
    download_dvl_parser.add_argument('-f', '--fid', help='id of file to download')
    download_dvl_parser.add_argument('-s', '--sa', help='start byte address of file to download')
    download_dvl_parser.add_argument('-l', '--length', help='length of bytes in file to download')
    download_dvl_parser.add_argument('-p', '--path', help='path for file to be downloaded', completer=cmd2.Cmd.path_complete)

    @with_argparser(download_dvl_parser)
    @with_category(CMD_CAT_DOWNLOAD)
    def do_download_dvl_data(self, download_dvl_args):

        fid = download_dvl_args.fid
        sa = download_dvl_args.sa
        length = download_dvl_args.length
        path = download_dvl_args.path

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('Nucleus not connected')
            return

        if self.nucleus_driver.parser.thread.is_alive():
            self.nucleus_driver.messages.write_message('Can not download DVL diagnostics data while logging thread is running')
            return

        if fid is not None:
            fid = int(fid)

        if sa is not None:
            sa = int(sa)

        if length is not None:
            length = int(length)

        status = self.nucleus_driver.download_dvl_data(fid=fid, sa=sa, length=length, path=path)

        if status:
            self.nucleus_driver.messages.write_message('Successfully downloaded data')
        else:
            self.nucleus_driver.messages.write_warning('Failed to completely download data')

    download_nucleus_parser = Cmd2ArgumentParser(description='Download nucleus data')
    download_nucleus_parser.add_argument('-f', '--fid', help='id of file to download')
    download_nucleus_parser.add_argument('-s', '--sa', help='start byte address of file to download')
    download_nucleus_parser.add_argument('-l', '--length', help='length of bytes in file to download')
    download_nucleus_parser.add_argument('-p', '--path', help='path for file to be downloaded', completer=cmd2.Cmd.path_complete)

    @with_argparser(download_nucleus_parser)
    @with_category(CMD_CAT_DOWNLOAD)
    def do_download_nucleus_data(self, download_nucleus_args):

        fid = download_nucleus_args.fid
        sa = download_nucleus_args.sa
        length = download_nucleus_args.length
        path = download_nucleus_args.path

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('Nucleus not connected')
            return

        if self.nucleus_driver.parser.thread.is_alive():
            self.nucleus_driver.messages.write_message('Can not download Nucleus data while logging thread is running')
            return

        if fid is not None:
            fid = int(fid)

        if sa is not None:
            sa = int(sa)

        if length is not None:
            length = int(length)

        status = self.nucleus_driver.download_nucleus_data(fid=fid, sa=sa, length=length, path=path)

        if status:
            self.nucleus_driver.messages.write_message('Successfully downloaded data')
        else:
            self.nucleus_driver.messages.write_warning('Failed to completely download data')

    convert_nucleus_data_parser = Cmd2ArgumentParser(description='Nucleus binary file path')
    convert_nucleus_data_parser.add_argument('path', help='Select path for nucleus binary file', completer=cmd2.Cmd.path_complete)

    @with_argparser(convert_nucleus_data_parser)
    @with_category(CMD_CAT_DOWNLOAD)
    def do_convert_nucleus_data(self, convert_nucleus_data_args):

        path = convert_nucleus_data_args.path

        if self.nucleus_driver.parser.thread.is_alive():
            self.nucleus_driver.messages.write_message('Can not convert Nucleus data while logging thread is running')
            return

        status = self.nucleus_driver.convert_nucleus_data(path=path)

        if status:
            self.nucleus_driver.messages.write_message('Data conversion complete')
        else:
            self.nucleus_driver.messages.write_warning('Failed to convert data')

    list_files_parser = Cmd2ArgumentParser(description='List Nucleus or DVL files')
    list_files_parser.add_argument('data_type', choices=['nucleus', 'dvl', 'getall'], help='List Nucleus data or DVL diagnostics data')

    @with_argparser(list_files_parser)
    @with_category(CMD_CAT_DOWNLOAD)
    def do_list_files(self, list_files_args):

        data_type = list_files_args.data_type

        if not self.nucleus_driver.connection.get_connection_status():
            self.nucleus_driver.messages.write_message('Nucleus not connected')
            return

        if self.nucleus_driver.parser.thread.is_alive():
            self.nucleus_driver.messages.write_message('Can not get file list while logging thread is running')
            return

        src = None
        if data_type == 'nucleus':
            src = 0
        if data_type == 'dvl':
            src = 1
        if data_type == 'getall':
            src = 2

        response = self.nucleus_driver.commands.list_files(src=src)

        for entry in response:
            self.nucleus_driver.messages.write_message(entry)


def nucleus_driver_console():

    nucleus_driver = NucleusDriver()

    app = App(nucleus_driver=nucleus_driver)

    app.cmdloop()



import time
from binascii import crc32
import re
from zipfile import ZipFile


class Flash:

    def __init__(self, **kwargs):

        self.BIN_FILE_PATTERN = r'(\w+)-v(\d+).(\d+).(\d+)(.*)'
        self.LDR_FILE_PATTERN = r'(\w+)_V(\d+)_(\d+)'

        self.messages = kwargs.get('messages')
        self.connection = kwargs.get('connection')
        self.commands = kwargs.get('commands')

        self.nucleus_firmware_name = None
        self.dvl_firmware_name = None
        self.nucleus_firmware = None
        self.dvl_firmware = None

    def set_flash_files(self, path: str) -> bool:

        status = False

        if path.endswith('.zip'):

            archive = ZipFile(path, 'r')

            for index, file in enumerate(archive.namelist()):
                if file.endswith('.bin'):
                    self.nucleus_firmware_name = archive.namelist()[index]
                if file.endswith('.ldr'):
                    self.dvl_firmware_name = archive.namelist()[index]

            if self.nucleus_firmware_name is not None:
                self.nucleus_firmware = archive.read(self.nucleus_firmware_name)

                if self.nucleus_firmware[:4] != b'\x3d\xb8\xf3\x96':
                    self.messages.write_warning(message='Invalid Nucleus firmware')
                    self.nucleus_firmware = None
                    self.nucleus_firmware_name = None

            if self.dvl_firmware_name is not None:
                self.dvl_firmware = archive.read(self.dvl_firmware_name)

            status = True

        elif path.endswith('.bin'):

            self.nucleus_firmware_name = path.split('/')[-1]

            with open(path, mode='rb') as file:
                self.nucleus_firmware = file.read()

            if self.nucleus_firmware[:4] != b'\x3d\xb8\xf3\x96':
                self.messages.write_warning(message='Invalid Nucleus firmware')
                self.nucleus_firmware = None
                self.nucleus_firmware_name = None

            status = True

        elif path.endswith('.ldr'):

            self.dvl_firmware_name = path.split('/')[-1]

            with open(path, mode='rb') as file:
                self.dvl_firmware = file.read()

            status = True

        else:
            self.messages.write_warning('Invalid extension for flash file. Flash file not set')

        return status

    def reset_flash_files(self):

        self.nucleus_firmware_name = None
        self.dvl_firmware_name = None
        self.nucleus_firmware = None
        self.dvl_firmware = None

    def compare_firmware(self):

        nucleus_rematch = None
        dvl_rematch = None

        nucleus_firmware_match = None
        dvl_firmware_match = None

        firmware_version = dict()
        firmware_name_version = dict()

        for key in ['MAJOR', 'MINOR', 'PATCH', 'EXTRA', 'BUILD', 'DVLFW', 'DVLMINOR']:
            if key == 'EXTRA':
                firmware_name_version[key] = '""'
                firmware_version[key] = '""'
            elif key == 'BUILD':
                firmware_name_version[key] = '0'
                firmware_version[key] = None
            else:
                firmware_name_version[key] = None
                firmware_version[key] = None

        for i in range(5):
            reply = self.commands.get_fw(_nmea=True)

            if len(reply) == 2 and reply[1] == b'$PNOR,OK*2B\r\n':
                break

        else:
            self.messages.write_warning('was not able to retrieve firmware information')
            return False, False, ''

        get_fw = reply[0].lstrip(b'$PNOR,GETFW,').split(b'*')[0]
        get_fw_list = get_fw.decode().split(',')

        for entry in get_fw_list:
            try:
                data = entry.split('=')
                if data[0] in firmware_version.keys():
                    firmware_version[data[0]] = data[1]
            except Exception as e:
                self.messages.write_warning('Encountered an error when parsing get_fw data: {}'.format(e))

        if self.nucleus_firmware_name is not None:
            nucleus_rematch = re.match(self.BIN_FILE_PATTERN, self.nucleus_firmware_name)
            if nucleus_rematch is not None:
                firmware_name_version['MAJOR'] = nucleus_rematch[2]
                firmware_name_version['MINOR'] = nucleus_rematch[3]
                firmware_name_version['PATCH'] = nucleus_rematch[4]

                extra_build = nucleus_rematch[5].split('.')[0].split('-')

                for element in extra_build:
                    if 'internal' in element:
                        break

                    if element == '':
                        continue

                    try:
                        int(element)
                        firmware_name_version['BUILD'] = element
                        break
                    except ValueError:
                        firmware_name_version['EXTRA'] = '"{}"'.format(element)

        if self.dvl_firmware_name is not None:
            dvl_rematch = re.match(self.LDR_FILE_PATTERN, self.dvl_firmware_name)
            firmware_name_version['DVLFW'] = dvl_rematch[2]
            firmware_name_version['DVLMINOR'] = dvl_rematch[3]

        if self.nucleus_firmware_name is not None and not nucleus_rematch:
            self.messages.write_warning(message='Nucleus firmware does not contain version number in its file name')

        elif self.nucleus_firmware_name is not None:

            for key in ['MAJOR', 'MINOR', 'PATCH', 'EXTRA', 'BUILD']:
                if firmware_name_version[key] == firmware_version[key]:
                    nucleus_firmware_match = True
                else:
                    nucleus_firmware_match = False
                    break

        if self.dvl_firmware_name is not None and not dvl_rematch:
            self.messages.write_warning(message='DVL firmware does not contain version number in its file name')

        elif self.dvl_firmware_name is not None:

            for key in ['DVLFW', 'DVLMINOR']:
                if firmware_name_version[key] == firmware_version[key]:
                    dvl_firmware_match = True
                else:
                    dvl_firmware_match = False
                    break

        return nucleus_firmware_match, dvl_firmware_match, get_fw.decode()

    def pack_firmware(self, firmware):

        data_length = len(firmware)
        crc = crc32(firmware)
        firmware_package = bytearray(data_length.to_bytes(4, 'little'))
        firmware_package.extend(firmware)
        firmware_package.extend(crc.to_bytes(4, 'little'))

        return firmware_package

    def flash_firmware(self) -> int:

        if self.nucleus_firmware is None and self.dvl_firmware is None:
            self.messages.write_message(message='No firmware has been selected for flashing')
            return -1

        condition = 0

        nucleus_firmware_match, dvl_firmware_match = self.compare_firmware()[:2]

        if self.nucleus_firmware is not None and condition == 0 and nucleus_firmware_match is not True:
            condition = self.flash_nucleus_firmware()
        elif nucleus_firmware_match is True:
            self.messages.write_message(message='New Nucleus firmware matches old firmware. Nucleus flashing will not be executed')

        if self.dvl_firmware is not None and condition == 0 and dvl_firmware_match is not True:
            condition = self.flash_dvl_firmware()
        elif dvl_firmware_match is True:
            self.messages.write_message(message='New DVL firmware matches old firmware. DVL flashing will not be executed')

        if condition == 0:
            condition = self.confirm_firmware()

        if condition == 0:
            condition = self.set_default_values()

        return condition

    def flash_nucleus_firmware(self) -> int:

        if not self.connection.get_connection_status():
            self.messages.write_message(message='Nucleus not connected')
            return -101

        firmware_package = self.pack_firmware(self.nucleus_firmware)

        self.messages.write_message(message='Uploading Nucleus firmware...')

        reply = self.commands._upload(package=firmware_package)
        if reply is None:
            return -102
        elif b'ACK\r\n' not in reply:
            return -103
        elif b'OK\r\n' not in reply:
            return -104

        self.messages.write_message(message='Nucleus firmware uploaded')
        self.messages.write_message(message='Updating Nucleus firmware...')

        reply = self.commands._fw_update()
        if reply is None:
            return -105
        elif b'ACK\r\n' not in reply:
            return -106
        elif b'OK\r\n' not in reply:
            return -107

        self.messages.write_message(message='Nucleus firmware updated')
        self.messages.write_message(message='Restarting system, this could take up to 80 seconds...')

        if self.connection.get_connection_type() == 'serial':
            start_up_message = self.commands.read_start_up(timeout=80)
            if b'Nortek Fusion DVL1000\r\n' not in start_up_message[0] and b'Nortek Nucleus1000\r\n' not in start_up_message[0]:
                self.messages.write_message(message='System did not start up properly: {}'.format(start_up_message))
                return -108

            self.connection.get_info()

        elif self.connection.get_connection_type() == 'tcp':

            self.connection.disconnect()

            time.sleep(80)

            serial_number = self.connection.get_serial_number_from_tcp_hostname()

            if 'Fusion' in self.nucleus_firmware_name:
                condition = self.connection.set_tcp_hostname_from_serial_number(serial_number=serial_number, name='NortekFusion')
            else:
                condition = self.connection.set_tcp_hostname_from_serial_number(serial_number=serial_number, name='NORTEK')

            if condition is not True:
                self.messages.write_warning(message='Failed to retrieve hostname. Attempting to reconnect with old hostname')

            self.connection.connect(connection_type='tcp')

            if not self.connection.get_connection_status():
                self.messages.write_warning(message='Failed to reconnect to TCP after Nucleus flashing')
                return -109

        else:
            self.messages.write_warning('Unrecognized connection type')
            return -110

        self.messages.write_message(message='System started up properly')

        return 0

    def flash_dvl_firmware(self) -> int:

        # TODO: Figure out why DVL flashing is violently unstable

        if not self.connection.get_connection_status():
            self.messages.write_message(message='Nucleus not connected')
            return -201

        firmware_package = self.pack_firmware(self.dvl_firmware)

        self.messages.write_message(message='Uploading DVL firmware...')

        reply = self.commands._upload(package=firmware_package)

        if reply is None:
            return -202
        elif b'ACK\r\n' not in reply:
            return -203
        elif b'OK\r\n' not in reply:
            return -204

        self.messages.write_message(message='DVL firmware uploaded')

        reply = self.commands._dvl_update()
        if reply is None:
            return -205
        elif b'ACK\r\n' not in reply:
            return -206

        update_in_progress = True
        while update_in_progress:

            response = self.connection.readline(timeout=5)

            if b'OK\r\n' in response:
                self.messages.write_message(message='DVL firmware updated')
                break

            elif b'FAIL -1\r\n' in response:
                self.messages.write_exception(message='Received FAIL -1. CRC failure')
                return -207

            elif b'DVLUPDATE,PROGRESS=' in response:
                progress = response.rstrip(b'\r\n').split(b'=')[-1].decode()
                try:
                    if int(progress) % 10 == 0:
                        self.messages.write_message(message='{}%'.format(progress))
                except Exception as exception:
                    self.messages.write_warning(message='Failed to output DVL flashing progress: {}'.format(exception))

            else:
                self.messages.write_warning(message='received a message that was not "PROGRESS" or "OK": {}'.format(response))
                return -208

        return 0

    def set_default_values(self) -> int:

        if not self.connection.get_connection_status():
            self.messages.write_message(message='Nucleus not connected')
            return -301

        if b'OK\r\n' not in self.commands.set_default_config():
            self.messages.write_message(message='Did not receive OK when setting default config')
            return -302
        else:
            self.messages.write_message(message='default config values set')

        if b'OK\r\n' not in self.commands.save_config():
            self.messages.write_message(message='Did not receive OK when saving default config')
            return -303
        else:
            self.messages.write_message(message='default config values saved')

        if b'OK\r\n' not in self.commands.set_default_mission():
            self.messages.write_message(message='Did not receive OK when setting default mission')
            return -304
        else:
            self.messages.write_message(message='default mission values set')

        if b'OK\r\n' not in self.commands.save_mission():
            self.messages.write_message(message='Did not receive OK when saving default mission')
            return -305
        else:
            self.messages.write_message(message='default mission values saved')

        if b'OK\r\n' not in self.commands.set_default_magcal():
            self.messages.write_message(message='Did not receive OK when setting default magcal')
            return -306
        else:
            self.messages.write_message(message='default magcal values set')

        if b'OK\r\n' not in self.commands.save_magcal():
            self.messages.write_message(message='Did not receive OK when saving default magcal')
            return -307
        else:
            self.messages.write_message(message='default magcal values saved')

        return 0

    def confirm_firmware(self) -> int:

        if self.nucleus_firmware_name is None and self.dvl_firmware_name is None:
            return -401

        nucleus_firmware_match, dvl_firmware_match, version = self.compare_firmware()

        self.messages.write_message(message='New firmware version: {}'.format(version))

        if nucleus_firmware_match is not False and dvl_firmware_match is not False:
            reply = self.commands._fw_confirm()
            if b'OK\r\n' not in reply:
                return -402

            self.messages.write_message(message='New firmware confirmed')

        else:
            self.messages.write_warning(message='Firmware update failed. Nucleus did not start up properly')
            self.messages.write_warning(message='Reverting back to old firmware...')

            self.commands.reboot()

            if self.connection.get_connection_type() == 'serial':

                start_up_message = self.commands.read_start_up(timeout=100)

                if b'Nortek Fusion DVL1000\r\n' not in start_up_message[0] and b'Nortek Nucleus1000\r\n' not in start_up_message[0]:
                    self.messages.write_message(message='System did not start up properly')
                    return -403

            elif self.connection.get_connection_type == 'tcp':

                self.connection.disconnect()

                time.sleep(20)

                self.connection.connect(connection_type='tcp')

                if not self.connection.get_connection_status():
                    self.messages.write_warning(message='Failed to reconnect to TCP after Nucleus flashing')
                    return -404

            return -405

        self.messages.write_warning(message='Firmware update complete')

        return 0



from pathlib import Path
from binascii import crc32
from datetime import datetime


class Download:

    def __init__(self, **kwargs):

        self.messages = kwargs.get('messages')
        self.connection = kwargs.get('connection')
        self.commands = kwargs.get('commands')
        self.parser = kwargs.get('parser')
        self.logger = kwargs.get('logger')

        self._path = str(Path.cwd()) + '/download'

        self.dvl_download_statistics = dict()
        self.dvl_download_statistics['successful bytes'] = 0
        self.dvl_download_statistics['failed bytes'] = 0

        self.PACKET_LENGTH = 1024 * 10

    @staticmethod
    def _handle_crc(dvl_data: bytes):

        crc_fail_data = b''

        while True:
            crc_fail_data += dvl_data[:1]
            dvl_data = dvl_data[1:]

            if len(dvl_data) <= 1:
                return dvl_data, crc_fail_data

            elif len(dvl_data) <= 2 and dvl_data[:2] != b'\x00\x00' and int.from_bytes(dvl_data[:2], 'little') <= int.from_bytes(b'\x00\x40', 'little'):
                return dvl_data, crc_fail_data

            elif len(dvl_data) <= 3 and dvl_data[:2] != b'\x00\x00' and dvl_data[2:3] == b'\x00' and int.from_bytes(dvl_data[:2], 'little') <= int.from_bytes(b'\x00\x40', 'little'):
                return dvl_data, crc_fail_data

            elif dvl_data[:2] != b'\x00\x00' and dvl_data[2:4] == b'\x00\x00' and int.from_bytes(dvl_data[:2], 'little') <= int.from_bytes(b'\x00\x40', 'little'):
                return dvl_data, crc_fail_data

    @staticmethod
    def progress_bar(current, total, bar_length=20):
        fraction = current / total

        arrow = int(fraction * bar_length - 1) * '-' + '>'
        padding = int(bar_length - len(arrow)) * ' '

        ending = '\n' if current == total else '\r'

        print(f'Progress: [{arrow}{padding}] {int(fraction * 100)}%', end=ending)

    def set_path(self, path):

        self._path = path.rstrip('/') + '/download'

        return self._path

    def get_download_parameters(self, fid: int, src: int, sa: int, length: int) -> (bool, dict):

        def _get_file_list():

            if src not in [0, 1]:
                self.messages.write_warning('src argument must be 0 or 1')
                return False

            list_files = self.commands.list_files(src=src)

            if len(list_files) == 0 or list_files[-1] != b'OK\r\n':
                self.messages.write_warning('Not able to obtain file list from device')
                return False

            elif len(list_files) == 1 and list_files[0] == b'OK\r\n':
                self.messages.write_warning('No files available on device')
                return False

            # extract fids and lengths from file list
            for entry in list_files:
                if entry == b'OK\r\n':
                    break

                id = None
                for data in entry.split(b','):
                    if b'FID' in data:
                        try:
                            id = int(data.split(b'=')[1].decode())
                        except (ValueError, IndexError, AttributeError):
                            self.messages.write_warning('Failed to extract id from entry. Skipping this entry')
                            break

                    elif b'LEN' in data:
                        if id is not None:
                            try:
                                len_data = data.split(b'=')
                                fid_dict[id] = {'length': int(len_data[1])}
                            except (ValueError, IndexError, AttributeError):
                                self.messages.write_warning('Failed to extract len from entry. Skipping this entry')
                                break
                        else:
                            self.messages.write_warning('Found LEN before FID in download entry. entry not appended til list of available fids')
                            break

            if len(fid_dict.keys()) < 1:
                self.messages.write_warning('Not able to extract any entries from file list')
                return False

            return True

        def _get_download_parameters():

            if fid is None:
                try:
                    download_parameters['fid'] = max(fid_dict.keys())
                except (TypeError, ValueError):
                    self.messages.write_warning('Was not able to obtain latest fid from FID list')
                    return False

            elif fid in fid_dict.keys():
                download_parameters['fid'] = fid

            else:
                self.messages.write_warning('Could not find specified fid on device')
                return False

            if sa is None:
                download_parameters['sa'] = 0

            elif sa < fid_dict[download_parameters['fid']]['length']:
                download_parameters['sa'] = sa

            else:
                self.messages.write_warning('sa argument is larger or equal to length of specified fid')
                return False

            if length is None:
                download_parameters['end'] = fid_dict[download_parameters['fid']]['length']

            elif download_parameters['sa'] + length <= fid_dict[download_parameters['fid']]['length']:
                download_parameters['end'] = download_parameters['sa'] + length

            else:
                self.messages.write_warning('Specified sa and length argument is larger the file length. Setting end parameter to file length')
                download_parameters['end'] = fid_dict[download_parameters['fid']]['length']

            if download_parameters['fid'] is None or download_parameters['sa'] is None or download_parameters['end'] is None:
                self.messages.write_warning('Unable to specify all download parameters')
                return False

            return True

        fid_dict = dict()
        if not _get_file_list():
            return False, dict()

        download_parameters = dict()
        download_parameters['fid'] = None
        download_parameters['sa'] = None
        download_parameters['end'] = None
        if not _get_download_parameters():
            return False, dict()

        return True, download_parameters

    def download_data(self, fid, src, sa, length) -> (bool, bytes):

        expected_data_length = len(str(length)) + 2 + length + 2 + 10 + 4  # length of _len number + length of \r\n + length of _len + length of \r\n + length of crc and \r\n + length of OK\r\n

        command = b'DOWNLOAD,FID=' + str(fid).encode() + b',SRC=' + str(src).encode() + b',SA=' + str(sa).encode() + b',LEN=' + str(length).encode() + b',CRC=1\r\n'
        self.connection.write(command)
        data = self.connection.read(size=expected_data_length, timeout=3)
        self.commands._check_reply(data=data, terminator=b'OK\r\n', command=command)

        if len(data) != expected_data_length:
            self.messages.write_warning('received data from download reply is incorrect length')
            return False, b''

        length_reply = data[:len(str(length)) + 2]
        data_reply = data[len(str(length)) + 2:-14]
        crc_reply = data[-14:-4]
        ok_reply = data[-4:]

        if length_reply[-2:] != b'\r\n':
            self.messages.write_warning('unexpected format of the length reply from download command')
            return False, b''

        if data_reply[-2:] != b'\r\n':
            self.messages.write_warning('unexpected format of the data reply from download command')
            return False, b''

        if crc_reply[-2:] != b'\r\n':
            self.messages.write_warning('unexpected format of the crc reply from download command')
            return False, b''

        if ok_reply[-2:] != b'\r\n':
            self.messages.write_warning('unexpected format of the ok reply from download command')
            return False, b''

        try:
            crc = int(crc_reply[:-2].decode(), 16)
        except ValueError:
            self.messages.write_warning('Unable to convert received crc value to integer')
            return False, b''

        if crc32(data_reply[:-2]) == crc:
            return True, data_reply[:-2]
        else:
            return False, data_reply[:-2]

    def download_get_all_data(self, fid: int = None):

        def _extract_get_all(data):

            b = None
            b_prev = None
            c = data

            get_all = list()
            while True:
                if b is not None:
                    b_prev = b
                a, b, c = c.partition(b'\xa5')
                if b_prev is not None:
                    get_all.append(b_prev + a)
                if len(c) < 2:
                    break

            return get_all

        def _check_nucleus_data():

            if nucleus_data['sync bit'] != 0xa5:
                self.messages.write_warning('Incorrect "sync bit" for Nucleus get_all-data')
                return False

            if nucleus_data['header size'] != len(nucleus_data['header']):
                self.messages.write_warning('Incorrect "header size" for Nucleus get_all-data')
                return False

            if nucleus_data['data series id'] != 0xa0:
                self.messages.write_warning('Incorrect "data series id" for Nucleus get_all-data')
                return False

            if nucleus_data['family id'] != 0x20:
                self.messages.write_warning('Incorrect "family id" for Nucleus get_all-data')
                return False

            if nucleus_data['data size'] != len(nucleus_data['data']):
                self.messages.write_warning('"data size" does not match length of data for Nucleus get_all-data')
                return False

            if nucleus_data['data checksum'] != self.parser.checksum(nucleus_data['data']):
                self.messages.write_warning('"data checksum" does not match calculated crc of data for Nucleus get_all-data')
                return False

            if nucleus_data['header checksum'] != self.parser.checksum(nucleus_data['header'][:-2]):
                self.messages.write_warning('"header checksum" does not match calculated crc of data for Nucleus get_all-data')
                return False

            return True

        def _check_dvl_data():

            if dvl_data['sync bit'] != 0xa5:
                self.messages.write_warning('Incorrect "sync bit" for DVL get_all-data')
                return False

            if dvl_data['header size'] != len(dvl_data['header']):
                self.messages.write_warning('Incorrect "header size" for DVL get_all-data')
                return False

            if dvl_data['data series id'] != 0xa0:
                self.messages.write_warning('Incorrect "data series id" for DVL get_all-data')
                return False

            if dvl_data['family id'] != 0x16:
                self.messages.write_warning('Incorrect "family id" for DVL get_all-data')
                return False

            if dvl_data['data size'] != len(dvl_data['data'][:-2]):
                self.messages.write_warning('"data size" does not match length of data for DVL get_all-data')
                return False

            if dvl_data['data checksum'] != self.parser.checksum(dvl_data['data'][:-2]):
                self.messages.write_warning('"data checksum" does not match calculated crc of data for DVL get_all-data')
                return False

            if dvl_data['header checksum'] != self.parser.checksum(dvl_data['header'][:-2]):
                self.messages.write_warning('"header checksum" does not match calculated crc of data for DVL get_all-data')
                return False

            return True

        if fid is not None:
            try:
                int(fid)
            except ValueError:
                self.messages.write_warning('Invalid type for fid, must be integer')
                return None, None
            else:
                encoded_fid = b'FID=' + str(fid).encode() + b','
        else:
            encoded_fid = b''

        command = b'DOWNLOAD,' + encoded_fid + b'SRC=2\r\n'

        self.connection.write(command)
        get_all_data = self.commands._get_reply(terminator=b'OK\r\n', timeout=3)
        self.commands._check_reply(data=get_all_data, terminator=b'OK\r\n', command=command)

        if len(get_all_data) < 100 or get_all_data[-4:] != b'OK\r\n':
            self.messages.write_warning('Received incomplete data from attempting to download get_all data from device')
            return None, None

        get_all_list = _extract_get_all(get_all_data[:-4])

        if len(get_all_list) != 2:
            self.messages.write_warning('Failed to extract get_all data from download,src=2 reply')

            return None, None

        nucleus_data = dict()
        nucleus_data['sync bit'] = get_all_list[0][0]
        nucleus_data['header size'] = get_all_list[0][1]
        nucleus_data['data series id'] = get_all_list[0][2]
        nucleus_data['family id'] = get_all_list[0][3]
        nucleus_data['data size'] = get_all_list[0][4] | get_all_list[0][5] << 8
        nucleus_data['data checksum'] = get_all_list[0][6] | get_all_list[0][7] << 8
        nucleus_data['header checksum'] = get_all_list[0][8] | get_all_list[0][9] << 8
        nucleus_data['header'] = get_all_list[0][:10]
        nucleus_data['data'] = get_all_list[0][10:]

        dvl_data = dict()
        dvl_data['sync bit'] = get_all_list[1][0]
        dvl_data['header size'] = get_all_list[1][1]
        dvl_data['data series id'] = get_all_list[1][2]
        dvl_data['family id'] = get_all_list[1][3]
        dvl_data['data size'] = get_all_list[1][4] | get_all_list[1][5] << 8
        dvl_data['data checksum'] = get_all_list[1][6] | get_all_list[1][7] << 8
        dvl_data['header checksum'] = get_all_list[1][8] | get_all_list[1][9] << 8
        dvl_data['header'] = get_all_list[1][:10]
        dvl_data['data'] = get_all_list[1][10:]

        if _check_nucleus_data():
            get_all_nucleus = nucleus_data['data']
        else:
            get_all_nucleus = None

        if _check_dvl_data():
            get_all_dvl = get_all_list[1]
        else:
            get_all_dvl = None

        return get_all_nucleus, get_all_dvl

    def download_dvl_data(self, fid=None, sa=None, length=None, path=None) -> bool:

        def _check_arguments():

            if fid is not None and (not isinstance(fid, int) or fid < 1):
                self.messages.write_warning('fid argument must be a positive integer larger or equal to 1')
                return False

            if sa is not None and (not isinstance(sa, int) or sa < 0):
                self.messages.write_warning('sa argument must be a non-negative integer')
                return False

            if length is not None and (not isinstance(length, int) or length < 1):
                self.messages.write_warning('length argument must be a positive integer larger or equal to 1')
                return False

            return True

        def _download_data() -> (bool, bytes):

            failed_attempt = False
            for attempt in range(1, 11):
                status, package = self.download_data(fid=download_parameters['fid'], src=1, sa=index, length=min(download_parameters['end'] - index, self.PACKET_LENGTH))

                if status:
                    if failed_attempt:
                        self.messages.write_message('successfully received packet at index {}'.format(index))
                    break
                else:
                    failed_attempt = True
                    self.messages.write_message('Failed to receive packet at index {} on attempt {}. Retrying...'.format(index, attempt))

            else:
                self.messages.write_warning('Failed to receive package from DVL debug data download 10 consecutive attempts. Aborting download!')
                return False, b''

            return True, package

        def _download_get_all() -> (bool, bytes):

            for i in range(1, 11):
                get_all = self.download_get_all_data(download_parameters['fid'])[1]

                if get_all is not None:
                    break

            else:
                return False, b''

            return True, get_all

        def _write_data_to_file(_data: bytes, _file, _fail_file) -> bytes:

            previous_data = None
            while True:

                if previous_data == _data:
                    self.messages.write_exception('Encountered an unexected situation where previous data is equal current data in dvl file writing')
                    return _data
                else:
                    previous_data = _data

                if len(_data) < 4:
                    return _data

                length = int.from_bytes(_data[:4], 'little')

                if length > 10000:
                    _data, failed_data = self._handle_crc(_data)
                    _fail_file.write(failed_data)
                    self.dvl_download_statistics['failed bytes'] += len(failed_data)

                elif len(_data) >= 4 + length + 4:
                    dvl_data = _data[4:4 + length]
                    crc = int.from_bytes(_data[4 + length:4 + length + 4], 'little')

                    if crc == crc32(dvl_data):
                        _file.write(dvl_data)
                        _data = _data[4 + length + 4:]
                        self.dvl_download_statistics['successful bytes'] += len(dvl_data)
                    else:
                        _data, failed_data = self._handle_crc(_data)
                        _fail_file.write(failed_data)
                        self.dvl_download_statistics['failed bytes'] += len(failed_data)

                else:
                    return _data

        if not _check_arguments():
            return False

        status, download_parameters = self.get_download_parameters(src=1, fid=fid, sa=sa, length=length)
        if not status:
            return False

        status, get_all = _download_get_all()

        if status:

            self.dvl_download_statistics['successful bytes'] = 0
            self.dvl_download_statistics['failed bytes'] = 0
            percentage_previous = -1

            if path is None:
                path = self._path

            file_path = path.rstrip('/') + '/dvl/' + datetime.now().strftime('%y%m%d_%H%M%S')
            Path(file_path).mkdir(parents=True, exist_ok=True)

            self.messages.write_message('Downloading data to: {}'.format(file_path))

            with open(file_path + '/dvl_data.bin', 'wb') as file, open(file_path + '/dvl_crc_fails.bin', 'wb') as fail_file:

                file.write(get_all)

                downloaded_bytes = 0
                data = b''
                for index in range(download_parameters['sa'], download_parameters['end'], self.PACKET_LENGTH):

                    status, package = _download_data()

                    downloaded_bytes += len(package)
                    percentage = downloaded_bytes * 100 / (download_parameters['end'] - download_parameters['sa'])
                    if percentage > percentage_previous + 1:
                        if percentage > 99:
                            percentage = 100
                        self.progress_bar(percentage, 100)
                        percentage_previous = int(percentage)

                    if not status:
                        break

                    data += package

                    data = _write_data_to_file(data, file, fail_file)

                self.messages.write_message('Downloaded and converted {} bytes of data. {} bytes of data failed conversion due to CRC checks'.format(self.dvl_download_statistics['successful bytes'], self.dvl_download_statistics['failed bytes']))

        return status

    def download_nucleus_data(self, fid=None, sa=None, length=None, path=None) -> bool:

        def _check_arguments():

            if fid is not None and (not isinstance(fid, int) or fid < 1):
                self.messages.write_warning('fid argument must be a positive integer larger or equal to 1')
                return False

            if sa is not None and (not isinstance(sa, int) or sa < 0):
                self.messages.write_warning('sa argument must be a non-negative integer')
                return False

            if length is not None and (not isinstance(length, int) or length < 1):
                self.messages.write_warning('length argument must be a positive integer larger or equal to 1')
                return False

            return True

        def _download_data() -> (bool, bytes):

            failed_attempt = False
            for attempt in range(1, 11):
                status, package = self.download_data(fid=download_parameters['fid'], src=0, sa=index, length=min(download_parameters['end'] - index, self.PACKET_LENGTH))

                if status:
                    if failed_attempt:
                        self.messages.write_message('successfully received packet at index {}'.format(index))
                    break
                else:
                    failed_attempt = True
                    self.messages.write_message('Failed to receive packet at index {} on attempt {}. Retrying...'.format(index, attempt))

            else:
                self.messages.write_warning('Failed to receive package from DVL debug data download 10 consecutive attempts. Aborting download!')
                return False, b''

            return True, package

        def _download_get_all() -> (bool, bytes):

            for i in range(1, 11):
                get_all = self.download_get_all_data(download_parameters['fid'])[0]

                if get_all is not None:
                    break

            else:
                return False, b''

            return True, get_all

        if not _check_arguments():
            return False

        status, download_parameters = self.get_download_parameters(src=0, fid=fid, sa=sa, length=length)
        if not status:
            return False

        status, get_all = _download_get_all()

        if status:

            if path is None:
                path = self._path

            file_path = path.rstrip('/') + '/nucleus/' + datetime.now().strftime('%y%m%d_%H%M%S')
            Path(file_path).mkdir(parents=True, exist_ok=True)

            self.messages.write_message('Downloading data to: {}'.format(file_path))

            with open(file_path + '/get_all.txt', 'w') as file:
                file.writelines(get_all.decode())

            with open(file_path + '/nucleus_data.bin', 'wb') as file:
                file.write(get_all)
                percentage_previous = -1
                downloaded_bytes = 0
                for index in range(download_parameters['sa'], download_parameters['end'], self.PACKET_LENGTH):

                    status, package = _download_data()

                    downloaded_bytes += len(package)
                    percentage = downloaded_bytes * 100 / (download_parameters['end'] - download_parameters['sa'])
                    if percentage > percentage_previous + 1:
                        if percentage > 99:
                            percentage = 100
                        self.progress_bar(percentage, 100)
                        percentage_previous = int(percentage)

                    if not status:
                        break

                    file.write(package)

        return status

    def convert_nucleus_data(self, path) -> bool:

        if not Path(path).is_file():
            self.messages.write_warning('selected file path is not a file')
            return False

        path_folder = Path(path).parent

        self.logger.set_path(path=str(path_folder) + '/nucleus_converted')

        self.logger.start(_converting=True)

        with open(path, 'rb') as raw_file:

            raw_file.seek(0, 2)
            file_length = raw_file.tell()
            raw_file.seek(0, 0)

            percentage_previous = -1

            while True:
                data = raw_file.read(1)

                percentage = raw_file.tell() * 100 / file_length
                if percentage > percentage_previous + 1:
                    if percentage > 99:
                        percentage = 100
                    self.progress_bar(percentage, 100)
                    percentage_previous = int(percentage)

                if data == b'':
                    break

                self.parser.add_data(data)

            self.logger.stop()

        return True
